library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

use work.am2901_comp.all;
use work.am2901_mnemonics.all;

entity alu is

  generic (
    WIDTH : integer := 4);

  port (
    R_in         : in  std_logic_vector((WIDTH - 1) downto 0);
    S_in         : in  std_logic_vector((WIDTH - 1) downto 0);
    CI_in        : in  std_logic;
    ALU_Cntrl_in : in  std_logic_vector(2 downto 0);
    F_out        : out std_logic_vector((WIDTH - 1) downto 0);
    G_out        : out std_logic;
    P_out        : out std_logic;
    CO_out       : out std_logic;
    F_MSB_out    : out std_logic;
    OVR_out      : out std_logic;
    Feq0_out     : out std_logic);

end alu;

architecture arch_alu of alu is

  signal R         : std_logic_vector((WIDTH - 1) downto 0);
  signal S         : std_logic_vector((WIDTH - 1) downto 0);
  signal CI        : std_logic;
  signal ALU_Cntrl : std_logic_vector(2 downto 0);
  signal F         : std_logic_vector((WIDTH - 1) downto 0);
  signal G         : std_logic;
  signal P         : std_logic;
  signal CO        : std_logic;
  signal F_MSB     : std_logic;
  signal OVR       : std_logic;
  signal Feq0      : std_logic;

  signal Carry_P : std_logic_vector((WIDTH - 1) downto 0);
  signal Carry_G : std_logic_vector((WIDTH - 1) downto 0);
  signal Carry   : std_logic_vector(WIDTH downto (WIDTH - 1));

begin  -- arch_alu

  R         <= R_in;
  S         <= S_in;
  CI        <= CI_in;
  ALU_Cntrl <= ALU_Cntrl_in;

  F_out     <= F;
  G_out     <= G;
  P_out     <= P;
  CO_out    <= CO;
  F_MSB_out <= F_MSB;
  OVR_out   <= OVR;
  Feq0_out  <= Feq0;

  F_MSB <= F(WIDTH - 1);
  CO    <= Carry(WIDTH);

  F_Proc : process (ALU_Cntrl, R, S, CI)
  begin  -- process F_Proc

    case alu_cntrl is

      when ALU_ADD =>

        F <= R + S + CI;

      when ALU_SUBR =>

        F <= S - R - CI;

      when ALU_SUBS =>

        F <= R - S - CI;

      when ALU_OR =>

        F <= R or S;

      when ALU_AND =>

        F <= R and S;

      when ALU_NOTRS =>

        F <= (not R and S);

      when ALU_EXOR =>

        F <= R xor S;

      when ALU_EXNOR =>

        F <= not (R xor S);

      when others => null;

    end case;

  end process F_Proc;

  Feq0_Proc : process (F)

    variable Feq0_Prod : std_logic;

  begin  -- process Feq0_Proc

    Feq0_Prod := '0';

    for Feq_index in (WIDTH - 1) downto 0 loop

      Feq0_Prod := Feq0_Prod or F(Feq_index);

    end loop;  -- Feq_index

    Feq0 <= not Feq0_Prod;

  end process Feq0_Proc;

  Carry_PG_Proc : process (ALU_Cntrl, R, S)

    variable Carry_P_Sum  : std_logic_vector((WIDTH - 1) downto 0);
    variable Carry_G_Prod : std_logic_vector((WIDTH - 1) downto 0);

  begin  -- process Carry_PG_Proc

    case ALU_Cntrl is

      when ALU_ADD =>

        for ADD_index in (WIDTH - 1) downto 0 loop

          Carry_P_Sum(ADD_index)  := R(ADD_index) or S(ADD_index);
          Carry_G_Prod(ADD_index) := R(ADD_index) and S(ADD_index);

        end loop;  -- ADD_index

      when ALU_SUBR =>

        for SUBR_index in (WIDTH - 1) downto 0 loop

          Carry_P_Sum(SUBR_index)  := not R(SUBR_index) or S(SUBR_index);
          Carry_G_Prod(SUBR_index) := not R(SUBR_index) and S(SUBR_index);

        end loop;  -- SUBR_index

      when ALU_SUBS =>

        for SUBS_index in (WIDTH - 1) downto 0 loop

          Carry_P_Sum(SUBS_index)  := not R(SUBS_index) or S(SUBS_index);
          Carry_G_Prod(SUBS_index) := not R(SUBS_index) and S(SUBS_index);

        end loop;  -- SUBS_index

      when ALU_OR | ALU_AND | ALU_NOTRS | ALU_EXOR | ALU_EXNOR =>

        for index in (WIDTH - 1) downto 0 loop

          Carry_P_Sum(index)  := '0';
          Carry_G_Prod(index) := '0';

        end loop;  -- index

      when others => null;

    end case;  -- case ALU_Cntrl

    for index in (WIDTH - 1) downto 0 loop

      Carry_P(index) <= Carry_P_Sum(index);
      Carry_G(index) <= Carry_G_Prod(index);

    end loop;  -- index

  end process Carry_PG_Proc;

  Carry_Proc : process (ALU_Cntrl, Carry_P, Carry_G, CI)

    variable P_Prod       : std_logic_vector((WIDTH - 1) downto 0);
    variable P_Prod2      : std_logic_vector((WIDTH - 1) downto 0);
    variable G_Sum        : std_logic;
    variable Carry_W_Sum  : std_logic;
    variable Carry_W1_Sum : std_logic;

  begin  -- process Carry_Proc

    P_Prod(WIDTH - 1)  := Carry_P(WIDTH - 1);
    P_Prod2(WIDTH - 1) := '1';

    for n in (WIDTH - 2) downto 0 loop

      P_Prod(n)  := P_Prod(n + 1) and Carry_P(n);
      P_Prod2(n) := P_Prod2(n + 1) and Carry_P(n);

    end loop;  -- n    

    case ALU_Cntrl is
      when ALU_ADD | ALU_SUBR | ALU_SUBS =>

-------------------------------------------------------------------------------
        P <= P_Prod(0);
-------------------------------------------------------------------------------
        G_Sum := '0';
        for n in (WIDTH - 2) downto 0 loop

          G_Sum := G_Sum or (Carry_G(n) and P_Prod(n + 1));

        end loop;  -- n

        G <= Carry_G(WIDTH - 1) or G_Sum;
-------------------------------------------------------------------------------
        Carry_W1_Sum := '0';
        for n in (WIDTH - 2) downto 0 loop

          Carry_W1_Sum := Carry_W1_Sum or
                          (Carry_G(n) and P_Prod2(n + 1));
        end loop;  -- n

        Carry(WIDTH - 1) <= Carry_G(WIDTH - 2) or
                            Carry_W1_Sum or
                            (P_Prod2(0) and CI);
-------------------------------------------------------------------------------
        Carry_W_Sum := '0';
        for n in (WIDTH - 2) downto 0 loop

          Carry_W_Sum := Carry_W_Sum or
                         (Carry_G(n) and P_Prod(n + 1));

        end loop;  -- n

        Carry(WIDTH) <= Carry_G(WIDTH - 1) or
                        Carry_W_Sum or
                        (P_Prod(0) and CI);
-------------------------------------------------------------------------------
        OVR          <= Carry(WIDTH) xor Carry(WIDTH - 1);
-------------------------------------------------------------------------------

      when ALU_OR | ALU_AND | ALU_NOTRS | ALU_EXOR | ALU_EXNOR =>

        P                <= '1';
        G                <= '1';
        Carry(WIDTH - 1) <= '0';
        Carry(WIDTH)     <= '0';
        OVR              <= '0';

      when others => null;

    end case;  -- case ALU_Cntrl

  end process Carry_Proc;

end arch_alu;
