--
--      Name: Am2910 Microsequencer (am2910.vhd)
--
--      Description: This design implements an Am2910 microsequencer
--      which is used in complex state machine designs.
--
--      Designer: Thomas Mitchell
--
--      Revision History: 2 May 2005 - Initial version (TDM)
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;

use work.am2910_comp.all;

entity am2910 is
  generic (
    STACK_DEPTH  :     integer := 9);
  port (
    CP_in        : in  std_logic;       -- clock input
    nCC_in       : in  std_logic;       -- condition code
    nCCEN_in     : in  std_logic;       -- condition code enable
    CI_in        : in  std_logic;       -- carry in (to incrementer)
    D_in         : in  std_logic_vector(11 downto 0);  -- direct input
    I_in         : in  std_logic_vector(3 downto 0);  -- instruction
    nOE_in       : in  std_logic;       -- output enable for y output
    nRLD_in      : in  std_logic;       -- register/counter load enable
    nMap_En_out  : out std_logic;       -- mapping prom output enable
    nPL_En_out   : out std_logic;       -- pipeline register output enable
    nVect_En_out : out std_logic;       -- vector output enable
    Y_out        : out std_logic_vector(11 downto 0);  -- microprogram address
    nFull_out    : out std_logic);      -- stack full flag
end am2910;

architecture arch_am2910 of am2910 is

-- Internal signals

  signal CP         : std_logic;
  signal cc         : std_logic;
  signal ccen       : std_logic;
  signal CI         : std_logic;
  signal D          : std_logic_vector(11 downto 0);
  signal I          : std_logic_vector(3 downto 0);
  signal OE         : std_logic;
  signal Zero       : std_logic;
  signal RLD        : std_logic;
  signal Map_En     : std_logic;
  signal PL_En      : std_logic;
  signal Vect_En    : std_logic;
  signal Y          : std_logic_vector(11 downto 0);
  signal Full       : std_logic;
  signal R          : std_logic_vector(11 downto 0);
  signal F          : std_logic_vector(11 downto 0);
  signal Reg_Decr   : std_logic;
  signal Reg_Load   : std_logic;
  signal Stk_Clr    : std_logic;
  signal Stk_Push   : std_logic;
  signal Stk_Pop    : std_logic;
  signal Mux_Sel    : std_logic_vector(1 downto 0);
  signal Mux_En     : std_logic;
  signal PC         : std_logic_vector(11 downto 0);
  signal uPC        : std_logic_vector(11 downto 0);
  signal uProg_Addr : std_logic_vector(11 downto 0);

begin

-- Continuous assignments (inputs)

  CP <= CP_in;
  CI <= CI_in;
  D  <= D_in;
  I  <= I_in;

-- Inversion of active low inputs

  CC   <= not nCC_in;
  CCEN <= not nCCEN_in;
  OE   <= not nOE_in;
  RLD  <= not nRLD_in;

-- Continuous assignments (outputs)

  Y_out <= Y;

-- Inversion of active low outputs

  nMap_en_out  <= not Map_en;
  nPL_en_out   <= not PL_En;
  nVect_En_out <= not Vect_En;
  nFull_out    <= not Full;

-- Component Instantiations

  U1 : reg_count
    port map(
      Clock_in       => CP,
      Direct_Load_in => RLD,
      Decr_En_in     => Reg_Decr,
      Load_En_in     => Reg_Load,
      D_in           => D,
      Q_out          => R);

  U2 : zero_detector
    port map(
      D_in     => R,
      Zero_out => Zero);

  U3 : stack
    generic map (
      DEPTH     => STACK_DEPTH)
    port map(
      Clock_in  => CP,
      Stack_in  => PC,
      Clr_in    => Stk_Clr,
      Push_in   => Stk_Push,
      Pop_in    => Stk_Pop,
      Empty_out => open,
      Full_out  => Full,
      Stack_out => F);

  U4 : multiplexer
    port map(
      D0_in      => PC,
      D1_in      => F,
      D2_in      => R,
      D3_in      => D,
      Mux_Sel_in => Mux_Sel,
      Mux_En_in  => Mux_En,
      Mux_out    => uProg_Addr);

  U5 : program_counter
    port map(
      Clock_in => CP,
      D_in     => uPC,
      Q_out    => PC);

  U6 : incrementer
    port map(
      A_in     => uProg_Addr,
      Carry_in => CI,
      Sum_out  => uPC);

  U7 : instruction_pla
    port map(
      Cond_Code_in    => CC,
      Cond_Code_En_in => CCEN,
      Instr_in        => I,
      RegZero_in      => Zero,
      Mux_Sel_out     => Mux_Sel,
      Mux_En_out      => Mux_En,
      Reg_Decr_out    => Reg_Decr,
      Reg_Load_out    => Reg_Load,
      Stk_Clr_out     => Stk_Clr,
      Stk_Push_out    => Stk_Push,
      Stk_Pop_out     => Stk_Pop,
      PL_En_out       => PL_En,
      Map_En_out      => Map_En,
      Vect_En_out     => Vect_En);

  U8 : process(OE, uProg_Addr)
  begin

    if OE = '1' then

      Y <= uProg_Addr;

    else

      Y <= (others => '1');

    end if;  -- if OE = '1'

  end process U8;

end arch_am2910;

